#include <gtk/gtk.h>
#include <stdio.h>
#include "../Misc/my_dbug.h"
#include "../Misc/my_main.h"
#include "../Misc/mem_management.h" //for dbugging only
#include "../Misc/utilities.h"
#include "../Misc/error_interface.h"
#include "../Typechecker/typechecker.h"
#include "../Evaluator/evaluator.h"
//#include "../Evaluator/eval_simplify.h" //for dbugging only
//#include "../Evaluator/eval_expand.h" //for dbugging only
#include "syntaxtree.h"
#include "st_io.h"





static void 
printf_Signatures(void)
{
  GList* tmp;
  gchar* signature;
  DBUG_ENTER("printf_Signatures");

  tmp = syntax_tree;
  while (tmp != NULL) {
    signature = funNode2Signature(tmp->data);
    printf("%s\n", signature);
    mtfree(signature);
    tmp = g_list_next(tmp);
  }

  DBUG_VOID_RETURN;
} //printf_Signatures


static void 
printf_menu(gint mode, gint nrMaxEvaluations)
{
  DBUG_ENTER("printf_menu");

  printf("\n(x) exit    (e) eval %i step(s)    (s) set nr of steps"
	 "\n(m) toggle mode [%i] <0=lmim, 1=lmom, 2=full>\t:", 
	 nrMaxEvaluations, mode);

  DBUG_VOID_RETURN;
} //printf_menu


static void 
tui(Expr_node_ptr tmp, Expr_node_ptr exprtree)
{
  gboolean stop = FALSE;
  gint     nrMaxEvaluations = 1;
  gint     mode = e_lmim, number;
  gchar    inputkey[4] = {0,0,0,0};

  DBUG_ENTER("tui");

  while (!stop) {
    printf_menu(mode, nrMaxEvaluations);
    scanf("%1s", inputkey);
    switch (inputkey[0]) {
    case 'x': 
      stop = TRUE;
      break;

    case 'e':
      if (FALSE == evaluate(tmp, exprtree, mode, nrMaxEvaluations, FALSE)) {
	printf_ErrorDescription();
	clear_ErrorDescription();
	stop = TRUE;
      }
      exprtree = tmp->child[0];
      break;

    case 's':
      printf("\nEnter nr of steps: ");
      scanf("%i", &nrMaxEvaluations);
      break;

    case 'm':
      mode = (mode + 1) % 3;
      break;

    default:
      if (1 == sscanf(inputkey, "%d",&number)) { 
	if (FALSE == evaluate(tmp, exprtree, mode, number, FALSE)) {
	  printf_ErrorDescription();
	  clear_ErrorDescription();
	  stop = TRUE;
	}
	exprtree = tmp->child[0];
      }
    } //switch
  } //while

  DBUG_VOID_RETURN;
} //tui


int MAIN
{
  gint res = 0;
  gchar* infilename = NULL;
  gchar instring[265];
  gchar* tmpStr;
  Expr_node_ptr exprtree = NULL;
  Expr_node_ptr tmp = NULL;

  DBUG_ENTER("main");

  BEGIN_ARGUMENTS
    DBUG_OPTIONS
    LONG_OPTIONS
    SHORT_OPTIONS
    NOT_A_OPTION { infilename = _actarg; }
  END_ARGUMENTS

  if (infilename == NULL) {
    infilename = "example.prg";
  }

  res =  readPrgFromFile(infilename);
  if (res == 0) {
    printf("Input accepted as a KIEL-ML programm.\n");
  } else {
    printf("Input is not a KIEL-ML programm\n");
    clear_ErrorDescription();
    mtshowtotals("error final");
  }

  if (res == 0) {
    nav_syntaxtree(syntax_tree);

    printf_Signatures();
    sprintf(instring, "0123456789");
    printf("Enter an expression:\n");
    //scanf("%264s", instring);
    fgets(instring, 265, stdin);
    printf("You entered: '%s'\n", instring);

    exprtree = readExprFromString(instring);
    //nav_expr_node(exprtree);

    if (exprtree != NULL) {
      printf("expression-string is valid.\n");
      //nav_expr_node(exprtree);

      tmpStr = exprNode2String(exprtree);
      printf("expr2String : '%s'\n", tmpStr);
      mtmultifree(1, tmpStr);

      tmp = create_dummy(); //just a dummy
      tmp->child[0] = exprtree;

      tui(tmp, exprtree);

      //clean up
      exprtree = tmp->child[0];
      tmp->child[0] = NULL;
      free_expr_node(tmp);

      free_expr_node(exprtree);
    } else {
      printf("expression-string is invalid.\n");
      clear_ErrorDescription();
    }

    { /**** dbug_only ****/
      /* nav_syntaxtree(syntax_tree); */
      //tree_parts should be NULL
      printf("Nr. of tree_parts-elms is %d\n", g_list_length(tree_parts));
      printf_treeparts(tree_parts);
      free_syntaxtree(syntax_tree);
      free_symboltable(symbol_table);
      mtshowtotals("after free_syntax & -symbol");
      printf_mtslots();
    }
  }

  DBUG_RETURN(res);
} //main
