%{

/*
 * A yacc-file to parse a KIEL-ML program file.
 */
/*# SED-BEGIN delete region #*/
  /* only a test */
/*# SED-END delete region #*/
#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include "../Misc/my_dbug.h"
#include "../Misc/error_interface.h"
#include "../Misc/garbage_collection.h"
#include "resources.h"
#include "symboltable.h"
#include "syntaxtree.h"
#include "prg.h"

extern FILE* yyin;
/*static char* yyyinfilename = NULL;*/


/* some things to avoid compiler warnings */
extern void yyy_InitScanner(void);
extern int yylex();
int yyparse();


static int       next_free_type_index;
static int       nr_specialtypes;
static int       nr_params;
/*# SED-BEGIN delete region #*/
/* act_datatype_index may be set only by actions for 
 * 'datatypeAndId' and 'andAndId' */
static int                   act_datatype_index; 
static Fun_node_ptr          act_pFunnode;
static Fun_node_ptr          tmp_pFunnode;
/*# SED-END delete region #*/
#define MAX_STR  64
static int                   tmp_type_index;
static char                  tmp_str[MAX_STR];
static Symboltable_entry_ptr tmp_pStEntry;
static struct _Mlval_struct  tmp_mlval;
static Expr_node_ptr idnode; 
static Product_part_ptr product_part; 


 /******************************************************************/
 /**** Some functions to handle the 'undefined_symbols' easier. ****/
 /******************************************************************/

static GList*
addUndefSymbol(GList* table, struct _Strval_struct strval, gint datatype_index)
{
  DBUG_ENTER("addUndefSymbol");
  printf("l: %d,  c: %d, name: %s \n", strval.lineno, strval.colno,strval.str);

  DBUG_RETURN( addSymbol(undefined_symbols, strval.str, e_st_unknown, 
                         strval.lineno, datatype_index, 0,
                         (void*)strval.colno) );
}//addUndefSymbol


/*# SED-BEGIN delete region #*/
static gboolean 
isNewId(char* id)
{
  DBUG_ENTER("isNewId");

  DBUG_RETURN( seekSymbolId(symbol_table, id) == NULL );
} //isNewId


/** Tries to add a datatype symbol to the symbol_table and returnes the 
  * used type_index.
  * @param ref_nextfreetypeindex is a reference parameter to know the next
  *        unused type_index and to update it.
  * @return 0 in case of an error (redefinition) or the corresponding
  *         type_index to the datatype symbol.
  */
static gint
tryAddDatatypeSymbol(gchar* id, gint* ref_nextfreetypeindex)
{
  gint used_type_index;
  Symboltable_entry_ptr pSte;

  DBUG_ENTER("tryAddDatatypeSymbol");

  if (seekSymbolId(symbol_table, id) == NULL) {
    /* If this datatype has been used within a system of datatypes 
     * befor (than it is stored in 'undefined_symbols'), than there is 
     * allready a type_index assignet to this datatype. So check it out!
     */
    pSte = seekSymbolId(undefined_symbols, id);
    if (pSte != NULL) {
      //datatype allready in undefined_symbols
      used_type_index = SYMBOL_UNDEF_TYPE_INDEX(pSte);
    } else {
       //datatype never used befor, so get a new index
       used_type_index = *ref_nextfreetypeindex;
       *ref_nextfreetypeindex += 1;
    }
    //add this new symbol
    symbol_table = addSymbol(symbol_table, id, e_st_datatype, 0, 
                             used_type_index, 0, NULL);
  } else {
    used_type_index = 0;
  }

  DBUG_RETURN(used_type_index);
} //tryAddDatatypeSymbol


static gboolean
tryAddConstructorSymbol(gchar* id, gint nr_params, gint type_index)
{
  gboolean error = FALSE;

  DBUG_ENTER("tryAddConstructorSymbol");

  if (seekSymbolId(symbol_table, id) == NULL) {
    //add this new symbol
    symbol_table = addSymbol(symbol_table, id, e_st_constructor, nr_params, 
                             type_index, 0, NULL);
  } else {
    error = TRUE;
  }

  DBUG_RETURN(error);
} //tryAddConstructorSymbol

/*# SED-END delete region #*/

static gint
getTypeindexOfSpecialtype(struct _Strval_struct token, 
                          gint* ref_nextfreetypeindex)
{
  gint type_index;
  Symboltable_entry_ptr pSte;

  DBUG_ENTER("getTypeindexOfSpecialtype");

  //Is symbol allready defined?
  pSte = seekSymbolId(symbol_table, token.str);
  if (pSte != NULL) {
    type_index = pSte->type_index;
  } else {
    //Is symbol allready in 'undefined_symbols'?
    pSte = seekSymbolId(undefined_symbols, token.str);
    if (pSte != NULL) {
      //Symbol is undefined but allready in 'undefined_symbols'.
      type_index = SYMBOL_TYPE_INDEX(pSte);
    } else {
      //Symbol is undefined and never seen befor.
      type_index = *ref_nextfreetypeindex;
      *ref_nextfreetypeindex += 1;
      undefined_symbols = addUndefSymbol(undefined_symbols, token,
                                         type_index);
    }
    //printf_symboltable(undefined_symbols);
  }  //if

  DBUG_RETURN(type_index);
} //getTypeindexOfSpecialtype


/** SymboTable_ERROR
  */
static int 
st_error(char* msg, struct _Strval_struct strval)
{
  const int c_max = 80;
  char line0[c_max], line1[c_max], line2[c_max], line3[c_max], line4[c_max], 
       line5[c_max];

  DBUG_ENTER("st_error");

  DBUG_PRINT("", ("l:%.3d c:%.2d '%s' at '%s'\n", 
                  strval.lineno, strval.colno, msg, strval.str));
  sprintf(line0, TEXT_PROGRAMM_PARSE_FEHLER0, msg);
  sprintf(line1, TEXT_PROGRAMM_PARSE_FEHLER1, strval.lineno);
  sprintf(line2, TEXT_PROGRAMM_PARSE_FEHLER2, strval.colno);
  sprintf(line3, TEXT_PROGRAMM_PARSE_FEHLER3, "ID");
  sprintf(line4, TEXT_PROGRAMM_PARSE_FEHLER4, strval.str);
  sprintf(line5, TEXT_PROGRAMM_PARSE_FEHLER5);

  //check afterwards, if local variables lineX were big enough
  DBUG_ASSERT(strlen(line0) < c_max);
  DBUG_ASSERT(strlen(line1) < c_max);
  DBUG_ASSERT(strlen(line2) < c_max);
  DBUG_ASSERT(strlen(line3) < c_max);
  DBUG_ASSERT(strlen(line4) < c_max);
  DBUG_ASSERT(strlen(line5) < c_max);

  set_ErrorDescription(1, BUTTON_WARNUNG_OK_TEXT, line0, line1,
                       line2, line3, line4, line5);
  //clean up some things (e.g. free mem)
  yyyerror_cleanup();

  DBUG_RETURN(1);
} //st_error


/*# SED-BEGIN delete region #*/
/** TypeChecking_ERROR is a error handling function for this c-file 
  */
static void
differentClauses_error(gint lineno, gint colno, gchar* msg1, gchar* msg2)
{
  const gint c_max = 48;
  gchar line0[c_max], line1[c_max], line2[c_max], line3[c_max], line4[c_max], 
        line5[c_max];

  DBUG_ENTER("differentClauses_error");

  sprintf(line0, "Different clauses");
  sprintf(line1, "-----------------");
  sprintf(line2, "In line %d at pos %d ", lineno, colno);
  sprintf(line3, "%s", msg1);
  sprintf(line4, "%s", msg2);
  sprintf(line5, " ");

  //check afterwards, if local variables lineX were big enough
  DBUG_ASSERT(strlen(line0) < c_max);
  DBUG_ASSERT(strlen(line1) < c_max);
  DBUG_ASSERT(strlen(line2) < c_max);
  DBUG_ASSERT(strlen(line3) < c_max);
  DBUG_ASSERT(strlen(line4) < c_max);
  DBUG_ASSERT(strlen(line5) < c_max);

  set_ErrorDescription(1, BUTTON_WARNUNG_OK_TEXT, line0, line1,
                       line2, line3, line4, line5);
  //clean up so things (e.g. free mem)
  yyyerror_cleanup();

  DBUG_VOID_RETURN;
} //differentClauses_error
/*# SED-END delete region #*/


static int 
checkUndefinedSymbolsAgain(void)
{
  int res = 0;
  GList* tmp;
  Symboltable_entry_ptr pSte;
  struct _Strval_struct strval;

  DBUG_ENTER("checkUndefinedSymbolsAgain");
printf("\n\n\n\n");
printf("***************** UNDEF *****\n");
printf_symboltable(undefined_symbols);
printf("***************** DEF *****\n");

printf_symboltable(symbol_table);

  for (tmp = undefined_symbols;  ((tmp != NULL) && (res == 0)); 
       tmp = g_list_next(tmp))
  {
    pSte = tmp->data;
printf("checking undefined %s\n", pSte->id);
    if (seekSymbolId(symbol_table, pSte->id) == NULL) {
	printf("not found!\n");
      res = 1;
      strval.str    = SYMBOL_UNDEF_ID(pSte);
      strval.lineno = SYMBOL_UNDEF_LINENO(pSte);
      strval.colno  = (int)SYMBOL_UNDEF_COLNO(pSte);
      st_error("undefined symbol", strval);
    }
  } //for

  DBUG_RETURN(res);
} //checkUndefinedSymbolsAgain



%}

%union {
  struct _Noval_struct  noval;
  struct _Strval_struct strval;
  struct _Mlval_struct  mlval;
  Fun_node_ptr          fun_node;
  Expr_node_ptr         expr_node;
}

%token SCANNER_ERROR
%token <noval> COMMA COLON SEMICOLON BRA_L BRA_R SQRBRA_L SQRBRA_R VBAR 
%token <noval> UNDERSCORE
%token <noval> DATATYPE OF FUN AND IF THEN ELSE UNIT BOOL INT REAL STRING LIST LET IN END

%token <mlval> BOOL_CONST INT_CONST REAL_CONST STRING_CONST LIST_CONST 

%token <noval> UNARY_NOT UNARY_MINUS EQUAL UNEQUAL LT LE GT GE ANDALSO ORELSE 
%token <noval> PLUS MINUS MULT REAL_DIV INT_DIV MOD 
%token <noval> STRING_CONCAT PREPEND LIST_CONCAT 
%token <noval> ABS_ MIN_ MAX_ SQRT SIN COS EXP LN TRUNCATE SIZE ORD CHR PROJECT
%token <noval> SUBSTRING NULL_ HD TL LENGTH REV 

%token <strval> ID 

  /*# SED-BEGIN delete region #*/
%type <fun_node> fun_class
%type <expr_node> pattern pattern_list1 pattern_list2
%type <expr_node> param_listBRA_R special_type
  /*# SED-END delete region #*/
%type <strval> type
%type <expr_node> product_type constant expr infix_expr applbasicfun atomic_expr
%type <expr_node> arg_list expr_list letvars_listBRA_R 
  /*# SED-BEGIN insert region #
%type <expr_node> expr_root
  # SED-END insert region #*/

  /* Using the PREC_ELSE precedence (see rules for expr) and keeping
   * it lower than ORELSE and ANDALSO makes YACC handling 
   * 'if e then e else e orelse e' aquivalent to
   * 'if e then e else (e orelse e)'.
   * Per definition (see [Mic] chapter 2.13) ORELSE has lower precedence 
   * than ANDALSO. For other precedences see [Mic] appendix C.3.
   */
%nonassoc PREC_ELSE
%left     ORELSE
%left     ANDALSO
%left     EQUAL UNEQUAL LT LE GT GE
%right    PREPEND LIST_CONCAT
%left     PLUS MINUS STRING_CONCAT
%left     MULT INT_DIV MOD REAL_DIV
%nonassoc PREC_UNARY_MINUS PREC_UNARY_NOT

%start ml_program


%%


  /**************************************************************************
   * The following grammar is close to the grammar given in [Mic] (see
   * appendix B and C.3) , but restricted in some cases.
   **************************************************************************/

  /*# SED-BEGIN delete region #*/
ml_program : {} 
           | ml_program type_dec SEMICOLON { 
               /* type_dec adds all constructors to the syntay_tree */

	       //check if undefined_symbols are defined now
	       //printf("undefined datatype-symbols are:\n");
               //printf_symboltable(undefined_symbols);
               if (checkUndefinedSymbolsAgain()) {return(1);}
               undefined_symbols = free_symboltable(undefined_symbols);
             }
           | ml_program fun_dec SEMICOLON  {
               /* fun_dec adds all functions to the syntay_tree */

               //check if undefined_symbols are defined now
               //printf("undefined fun-symbols are:\n");
               //printf_symboltable(undefined_symbols);
               if (checkUndefinedSymbolsAgain()) {return(1);}
               undefined_symbols = free_symboltable(undefined_symbols);
             }
           ;

  /* type declarations */
type_dec : datatypeAndId EQUAL constructor_binding {
             /* constructor_binding adds all constructors to the syntay_tree */
           }
         | type_dec andAndId EQUAL constructor_binding {
           }
         ;

datatypeAndId : DATATYPE ID 
                { 
                  act_datatype_index = tryAddDatatypeSymbol($2.str, 
                                                        &next_free_type_index);
		  if (act_datatype_index == 0) {
                    st_error("redefined symbol", $2);
                    return(1);
                  } //if
                }
              ;

andAndId : AND ID 
           { 
             act_datatype_index = tryAddDatatypeSymbol($2.str,
						       &next_free_type_index);
	     if (act_datatype_index == 0) {
               st_error("redefined symbol", $2);
               return(1);
             } //if
           }
         ;

constructor_binding : ID {
                        if (TRUE == tryAddConstructorSymbol(
                                         $1.str, 
                                         0, 
                                         act_datatype_index)
                           ) {
                          st_error("redefined symbol", $1);
                          return(1);
			} //if
			tmp_pFunnode = create_fun_node($1.str, e_constructor, 
					 NULL, act_datatype_index, NULL, NULL);
                        syntax_tree = g_list_append(syntax_tree, tmp_pFunnode);
                        /* Now update the definition_point in the corresponding
			 * symbol_table_entry. */
                        tmp_pStEntry = seekSymbolId(symbol_table, $1.str);
                        DBUG_ASSERT(tmp_pStEntry != NULL);
                        tmp_pStEntry->definition_point = tmp_pFunnode;
                      }
                    | ID OF special_type {
                        if (TRUE == tryAddConstructorSymbol(
                                         $1.str, 
                                         nr_specialtypes, 
                                         act_datatype_index)
                           ) {
                          st_error("redefined symbol", $1);
                          return(1);
                        } //if
			tmp_pFunnode = create_fun_node($1.str, e_constructor, 
					   $3, act_datatype_index, NULL, NULL);
                        syntax_tree = g_list_append(syntax_tree, tmp_pFunnode);
                        /* Now update the definition_point in the corresponding
			 * symbol_table_entry. */
                        tmp_pStEntry = seekSymbolId(symbol_table, $1.str);
                        DBUG_ASSERT(tmp_pStEntry != NULL);
                        tmp_pStEntry->definition_point = tmp_pFunnode;
                      }
                    | constructor_binding VBAR ID {
                        if (TRUE == tryAddConstructorSymbol(
                                         $3.str, 
                                         0, 
                                         act_datatype_index)
                           ) {
                          st_error("redefined symbol", $3);
                          return(1);
                        } //if
			tmp_pFunnode = create_fun_node($3.str, e_constructor, 
					 NULL, act_datatype_index, NULL, NULL);
                        syntax_tree = g_list_append(syntax_tree, tmp_pFunnode);
                        /* Now update the definition_point in the corresponding
			 * symbol_table_entry. */
                        tmp_pStEntry = seekSymbolId(symbol_table, $3.str);
                        DBUG_ASSERT(tmp_pStEntry != NULL);
                        tmp_pStEntry->definition_point = tmp_pFunnode;
                      }
                    | constructor_binding VBAR ID OF special_type {
                        if (TRUE == tryAddConstructorSymbol(
                                         $3.str, 
                                         nr_specialtypes, 
                                         act_datatype_index)
                           ) {
                          st_error("redefined symbol", $3);
                          return(1);
                        } //if
			tmp_pFunnode = create_fun_node($3.str, e_constructor, 
					   $5, act_datatype_index, NULL, NULL);
                        syntax_tree = g_list_append(syntax_tree, tmp_pFunnode);
                        /* Now update the definition_point in the corresponding
			 * symbol_table_entry. */
                        tmp_pStEntry = seekSymbolId(symbol_table, $3.str);
                        DBUG_ASSERT(tmp_pStEntry != NULL);
                        tmp_pStEntry->definition_point = tmp_pFunnode;
                      }
                    ;

special_type : type {
                 nr_specialtypes = 1;
                 tmp_type_index = getTypeindexOfSpecialtype(
                              $1, &next_free_type_index);
                 DBUG_ASSERT(tmp_type_index > 0);
                 sprintf(tmp_str, "param%i", nr_specialtypes);
                 $$ = create_expr_param_list( 
                       create_expr_specialtype_param(tmp_str, tmp_type_index), 
                       NULL);
               }
             | type MULT special_type { 
                 nr_specialtypes++;
                 tmp_type_index = getTypeindexOfSpecialtype(
                              $1, &next_free_type_index);
                 DBUG_ASSERT(tmp_type_index > 0);
                 sprintf(tmp_str, "param%i", nr_specialtypes);
                 printf("neu: param%i\n", nr_specialtypes);
                 $$ = create_expr_param_list( 
                       create_expr_specialtype_param(tmp_str, tmp_type_index), 
                       $3);
	       }
             ;
        /*# SED-END delete region #*/

product_type : type {
                 nr_specialtypes = 1;
                 tmp_type_index = getTypeindexOfSpecialtype(
                              $1, &next_free_type_index);
                 DBUG_ASSERT(tmp_type_index > 0);
                 sprintf(tmp_str, $1.str);
                 $$ = create_expr_param_list( 
                       create_expr_producttype_param(tmp_str, tmp_type_index), 
                       NULL);
		 product_part = createProduct_part(tmp_type_index);

               }
             | type MULT product_type { 
                 nr_specialtypes++;
                 tmp_type_index = getTypeindexOfSpecialtype(
                              $1, &next_free_type_index);
                 DBUG_ASSERT(tmp_type_index > 0);
                 sprintf(tmp_str, "%s * %s", $1.str, ARG_EXPR_NODE($3)->strrep);
                 printf("produkttyp: %s mit id %i\n", tmp_str,tmp_type_index);
                 $$ = create_expr_param_list( 
                       create_expr_producttype_param(tmp_str, tmp_type_index), 
                       $3);
		 product_part = prependProduct_part(product_part,tmp_type_index);
	       }
             ;



type : UNIT      { $$.str = "unit"; 
                   $$.lineno = $1.lineno;  $$.colno = $1.colno; }
     | BOOL      { $$.str = "bool"; 
                   $$.lineno = $1.lineno;  $$.colno = $1.colno; }
     | INT       { $$.str = "int"; 
                   $$.lineno = $1.lineno;  $$.colno = $1.colno; }
     | REAL      { $$.str = "real"; 
                   $$.lineno = $1.lineno;  $$.colno = $1.colno; }
     | STRING    { $$.str = "string"; 
                   $$.lineno = $1.lineno;  $$.colno = $1.colno; }
     | ID        { $$ = $1; }
     | type LIST { 
                   tmp_pStEntry  = seekSymbolId(symbol_table, $1.str);
                   if (tmp_pStEntry != NULL) {
                     strncpy(tmp_str, $1.str, MAX_STR - 1);
                     strncat(tmp_str, " list", MAX_STR - strlen(tmp_str) - 1);
                     //do not use assignments like $$.str = ...
                     //use setYylval_strval instead (garbage collection)
                     setYylval_strval(&($$), tmp_str, $1.lineno, $1.colno);
                     tmp_type_index = tmp_pStEntry->type_index;
                     //store list-type to symboltable, if not allready done
                     tmp_pStEntry = seekSymbolId(symbol_table, $$.str);
                     if (tmp_pStEntry == NULL) {
                       symbol_table = addSymbol(symbol_table, 
                                 $$.str, e_st_basictype, 0,
                                 next_free_type_index, 
                                 tmp_type_index, NULL);
                       next_free_type_index++;
		     }
		   } else {
                     st_error("undefined symbol", $1); //undefined type
                     return(1);
		   }
                 }
           | product_type { 
     		printf("found product-type xxxxxx\n");
     		printf(ARG_EXPR_NODE($1)->strrep);
     		printf("\n");	
     		$$.str = ARG_EXPR_NODE($1)->strrep;//"int";
                
                tmp_pStEntry = seekSymbolId(symbol_table, $$.str);
                if (tmp_pStEntry == NULL) {
     		   printf("found product-type, adding with type_index %i type is product \n",next_free_type_index);
                   symbol_table = addSymbol(symbol_table, 
                            $$.str, e_st_product, 0,
                            next_free_type_index, 
                            tmp_type_index, NULL);
                   next_free_type_index++;
                   tmp_pStEntry = seekSymbolId(symbol_table, $$.str);
		   tmp_pStEntry->product_parts = product_part ;
		}
                tree_parts = myGlistRemove(tree_parts, $1);

               }
         
     ;

  /*# SED-BEGIN delete region #*/


  /* function declarations */
fun_dec : FUN fun_class         {syntax_tree = g_list_append(syntax_tree, $2);}
        | fun_dec AND fun_class {syntax_tree = g_list_append(syntax_tree, $3);}
        ;

fun_class : ID BRA_L param_listBRA_R COLON type EQUAL expr
            {
              if (isNewId($1.str)) {
                tmp_pStEntry  = seekSymbolId(symbol_table, $5.str);
                if (tmp_pStEntry != NULL) {
                  tmp_type_index = tmp_pStEntry->type_index;
                  act_pFunnode = create_fun_node($1.str, e_userfun, $3, 
                                            tmp_type_index, $7, binding_table);
	          //printf("bound id's of params for this fun are:\n");
                  //printf_symboltable(binding_table);
                  binding_table = NULL; //do not free here!!!
                  symbol_table = addSymbol(symbol_table, $1.str, e_st_userfun,
                                   nr_params, tmp_type_index, 0, act_pFunnode);
                } else {
                  st_error("undefined symbol", $5); //undefined resulttype
                  return(1);
                }  //if
	      } else {
                st_error("redefined symbol", $1); //fun_id allready defined
                return(1); 
              } //if
              $$ = act_pFunnode;
	    }
          | fun_class VBAR ID BRA_L param_listBRA_R COLON type EQUAL expr
            {
              /* Check if id and resulttype are the same as in the 
	       * first clause! */
	      if (0 != strcmp($3.str, act_pFunnode->fun_id)) {
		differentClauses_error($3.lineno, $3.colno, 
				       "The function ID is different to ", 
				       "the one in the first clause.");
		return 1;
	      } //if different ids
              tmp_pStEntry  = seekSymbolId(symbol_table, $7.str);
              if (tmp_pStEntry == NULL) {
                st_error("undefined symbol", $7); //undefined resulttype
                return(1);
	      } else {
		if (tmp_pStEntry->type_index != act_pFunnode->resulttype) {
		  differentClauses_error($7.lineno, $7.colno, 
					 "The RESULT-TYPE is different to ", 
					 "the one in the first clause.");
		  return 1;
		} //if different resulttypes
              }  //if undefined resulttype

	      /* add new clause to Fun_node */
	      append_fun_node(act_pFunnode, $5, binding_table, $9); 

	      //printf("bound id's of params for this fun are:\n");
	      //printf_symboltable(binding_table);
	      binding_table = NULL; //do not free here!!!
              $$ = act_pFunnode;
            }
          ;

  
param_listBRA_R : BRA_R {
               nr_params = 0;
               $$ = NULL;
             }
           | pattern COLON type BRA_R {
               nr_params = 0;
               tmp_pStEntry = seekSymbolId(symbol_table, $3.str);
               if (tmp_pStEntry != NULL) {
                 $1->valuetype_expected = tmp_pStEntry->type_index;
	       } else {
                 st_error("undefined symbol", $3); //undefined resulttype
                 return(1);
               }  //if
               $$ = create_expr_param_list($1, NULL);
             }
           | pattern COLON type COMMA param_listBRA_R {
               nr_params++;
               tmp_pStEntry = seekSymbolId(symbol_table, $3.str);
               if (tmp_pStEntry != NULL) {
                 $1->valuetype_expected = tmp_pStEntry->type_index;
	       } else {
                 st_error("undefined symbol", $3); //undefined resulttype
                 return(1);
               }  //if
               $$ = create_expr_param_list($1, $5);
             }
           ;



pattern : UNDERSCORE {
            $$ = create_expr_wildcard($1);
          }
        | constant {
            $$ = $1;
          }
        | ID { /* may be a variable or a parameterless constructor */ 
            tmp_pStEntry = seekSymbolId(symbol_table, $1.str);
            if (tmp_pStEntry != NULL) {
              if (tmp_pStEntry->symbol_type == e_st_constructor) {
                $$ = create_expr_constructor($1, NULL);
              } else {
                st_error("symbol is neither unique nor a constructor", $1);
                return(1);
	      } //if
	    } else {
              binding_table = addSymbol(binding_table, $1.str, e_st_param, 
	                                               0, 0, 0, NULL);
              $$ = create_expr_variable($1);
	    } //if
          }
        | BRA_L pattern BRA_R { 
            $$ = $2;
          }
        | SQRBRA_L SQRBRA_R { /* nil-list */
	    tmp_mlval.val.mllist = NULL;
	    tmp_mlval.lineno = $1.lineno;
	    tmp_mlval.colno  = $1.colno;
            //$$ = create_nil(tmp_mlval);
	    $$ = create_expr_listkeeper(tmp_mlval, NULL);
          }
        | SQRBRA_L pattern_list1 SQRBRA_R { /* mllist of a fixed size */
	    tmp_mlval.val.mllist = NULL;
	    tmp_mlval.lineno = $1.lineno;
	    tmp_mlval.colno  = $1.colno;
            //$$ = $2;
            $$ = create_expr_listkeeper(tmp_mlval, $2);
          }
        | pattern PREPEND pattern {
            $$ = create_expr_infixop($2, e_prepend, $1, $3);
          }
        | ID BRA_L pattern_list2 BRA_R { /* constructor with parameters */
            tmp_pStEntry = seekSymbolId(symbol_table, $1.str);
            if ((tmp_pStEntry != NULL) 
                && (tmp_pStEntry->symbol_type == e_st_constructor)) {
              $$ = create_expr_constructor($1, $3);
            } else {
              st_error("not a valid pattern", $1);
              return(1);
	    } //if
          }
        ;
  /* The listitems of this list have to have the same type_index. */
pattern_list1 : pattern {
               /* tmp_mlval.val.mllist = NULL;
		  tmp_mlval.lineno = LINENO_NODE($1);
	          tmp_mlval.colno  = COLNO_NODE( $1);
                  $$ = create_expr_list($1, create_nil(tmp_mlval));*/
		  $$ = create_expr_list($1, NULL);
                }
              | pattern COMMA pattern_list1 {
                  $$ = create_expr_list($1, $3);
                }
              ;


  /* The listitems of this list may have different type_indices. */
pattern_list2 : pattern {
                  $$ = create_expr_param_list($1, NULL);
                }
              | pattern COMMA pattern_list2 {
                  $$ = create_expr_param_list($1, $3);
                }
              ;
  /*# SED-END delete region #*/


constant : BRA_L BRA_R  { struct _Mlval_struct  mlval;
                          tmp_pStEntry = seekSymbolId(symbol_table, "unit");
                          DBUG_ASSERT(tmp_pStEntry != NULL);
                          mlval.val.mlunit = NULL;
                          mlval.lineno     = -1;
                          mlval.colno      = -1;
                          $$ = create_expr_constant(mlval,
                                                    tmp_pStEntry->type_index);
                        }
         | BOOL_CONST   { 
                          tmp_pStEntry = seekSymbolId(symbol_table, "bool");
                          DBUG_ASSERT(tmp_pStEntry != NULL);
                          $$ = create_expr_constant($1, 
                                                    tmp_pStEntry->type_index);
                        }
         | INT_CONST    { 
                          tmp_pStEntry = seekSymbolId(symbol_table, "int");
                          DBUG_ASSERT(tmp_pStEntry != NULL);
                          $$ = create_expr_constant($1, 
                                                    tmp_pStEntry->type_index);
                        }
         | REAL_CONST   { 
                          tmp_pStEntry = seekSymbolId(symbol_table, "real");
                          DBUG_ASSERT(tmp_pStEntry != NULL);
                          $$ = create_expr_constant($1, 
                                                    tmp_pStEntry->type_index);
                        }
         | STRING_CONST { 
                          tmp_pStEntry = seekSymbolId(symbol_table, "string");
                          DBUG_ASSERT(tmp_pStEntry != NULL);
                          $$ = create_expr_constant($1, 
                                                    tmp_pStEntry->type_index);
                        }
         | LIST_CONST   { //$$ = create_nil($1);
                          $$ = create_expr_listkeeper($1, NULL);
                        }
         ;

  /*# SED-BEGIN insert region #
expr_root: expr {
	     if (checkUndefinedSymbolsAgain()) {
printf("expr_root: checkUndefinedSymbolsAgain delivers an error\n");
               return(1);
             }
             undefined_symbols = free_symboltable(undefined_symbols);
	     $$ = $1;
           }
         | expr SEMICOLON {
	     if (checkUndefinedSymbolsAgain()) {
printf("expr_root: checkUndefinedSymbolsAgain delivers an error\n");
               return(1);
             }
             undefined_symbols = free_symboltable(undefined_symbols);
	     $$ = $1;
           }
         ;
  # SED-END insert region #*/


letvars_listBRA_R : BRA_R {
               nr_params = 0;
               $$ = NULL;
             }
           | ID COLON type BRA_R {

              
               tmp_pStEntry = seekSymbolId(symbol_table, $1.str);
               if (tmp_pStEntry != NULL) {
                   st_error("symbol is neither unique nor a constructor", $1);
                   return(1);
	       } else {
                 binding_table = addSymbol(binding_table, $1.str, e_st_param, 
	                                               0, 0, 0, NULL);
                 idnode = create_expr_variable($1);
                 idnode->subnodetype = e_boxed;
	       } //if

               nr_params = 0;
               tmp_pStEntry = seekSymbolId(symbol_table, $3.str);
               if (tmp_pStEntry != NULL) {
                 idnode->valuetype_expected = tmp_pStEntry->type_index;
              } else {
                 st_error("undefined symbol", $3); //undefined resulttype
                 return(1);
               }  //if


               $$ = create_expr_letvars_list(idnode, NULL);
             }
             
           | ID COLON type COMMA letvars_listBRA_R {
               nr_params++;
               
               tmp_pStEntry = seekSymbolId(symbol_table, $1.str);
               if (tmp_pStEntry != NULL) {
                   st_error("symbol is neither unique nor a constructor", $1);
                   return(1);
	       } else {
                 binding_table = addSymbol(binding_table, $1.str, e_st_param, 
	                                               0, 0, 0, NULL);
                 idnode = create_expr_variable($1);
                 idnode->subnodetype = e_boxed;
	       } //if

               tmp_pStEntry = seekSymbolId(symbol_table, $3.str);
               if (tmp_pStEntry != NULL) {
                 idnode->valuetype_expected = tmp_pStEntry->type_index;
	       } else {
                 st_error("undefined symbol", $3); //undefined resulttype
                 return(1);
               }  //if
               $$ = create_expr_letvars_list(idnode, $5);
             }

           ;


expr : 

    LET BRA_L letvars_listBRA_R  EQUAL expr IN expr END  {

         $$ = create_expr_let($3, $5, $7);
       }
| infix_expr        { $$ = $1; }
     | expr ANDALSO expr { $$ = create_expr_infixop($2, e_andalso, $1, $3); }
     | expr ORELSE expr  { $$ = create_expr_infixop($2, e_orelse, $1, $3); }
     | IF expr THEN expr ELSE expr %prec PREC_ELSE {
         $$ = create_expr_ifthenelse($1, $2, $4, $6);
       }
     | BRA_L expr_list BRA_R {
         $$ = create_expr_product($2);
     }
     ;

infix_expr : atomic_expr {}
           | UNARY_MINUS infix_expr %prec PREC_UNARY_MINUS {
             $$ = create_expr_unary_minus($1, $2);
             }
           | UNARY_NOT infix_expr %prec PREC_UNARY_NOT {
             $$ = create_expr_unary_not($1, $2);
             }
           | applbasicfun { /* These are applicaions of basic functions */
                            $$ = $1; 
             }
           | ID BRA_L BRA_R            
             { /* has to be a userfun? */
               tmp_pStEntry = seekSymbolId(symbol_table, $1.str);
               if (tmp_pStEntry == NULL) {
                 undefined_symbols = addUndefSymbol(undefined_symbols, $1, 0);
                 $$ = create_expr_userfun($1, NULL);
	       } else {
                 switch (tmp_pStEntry->symbol_type) {
                 case e_st_userfun: 
                   $$ = create_expr_userfun($1, NULL);
                   break;
                 default: st_error("id represents neither a function "
                                   "nor a constructor", $1);
                          return(1);
	         } //switch
	       } //if not in symboltable
             }
           | ID BRA_L arg_list BRA_R            
             { /* is it constructor or userfun? */
               tmp_pStEntry = seekSymbolId(symbol_table, $1.str);
               if (tmp_pStEntry == NULL) {
                 undefined_symbols = addUndefSymbol(undefined_symbols, $1, 0);
                 /* in this case it can't be a constructor, because undefined
                  * constructors may occur in a datatype-system only */
                 $$ = create_expr_userfun($1, $3);
	       } else {
                 switch (tmp_pStEntry->symbol_type) {
                 case e_st_constructor: 
                   $$ = create_expr_constructor($1, $3);
                   break;
                 case e_st_userfun: 
                   $$ = create_expr_userfun($1, $3);
                   break;
                 default: st_error("id represents neither a function "
                                   "nor a constructor", $1);
                          return(1);
	         } //switch
	       } //if not in symboltable
             }
           | infix_expr EQUAL         infix_expr
             { $$ = create_expr_infixop($2, e_equal, $1, $3); }
           | infix_expr UNEQUAL       infix_expr
             { $$ = create_expr_infixop($2, e_unequal, $1, $3); }
           | infix_expr LT            infix_expr
             { $$ = create_expr_infixop($2, e_lt, $1, $3); }
           | infix_expr LE            infix_expr
             { $$ = create_expr_infixop($2, e_le, $1, $3); }
           | infix_expr GT            infix_expr
             { $$ = create_expr_infixop($2, e_gt, $1, $3); }
           | infix_expr GE            infix_expr
             { $$ = create_expr_infixop($2, e_ge, $1, $3); }
           | infix_expr PLUS          infix_expr
             { $$ = create_expr_infixop($2, e_plus, $1, $3); }
           | infix_expr MINUS         infix_expr
             { $$ = create_expr_infixop($2, e_minus, $1, $3); }
           | infix_expr MULT          infix_expr
             { $$ = create_expr_infixop($2, e_mult, $1, $3); }
           | infix_expr MOD           infix_expr
             { $$ = create_expr_infixop($2, e_mod, $1, $3); }
           | infix_expr INT_DIV       infix_expr
             { $$ = create_expr_infixop($2, e_int_div, $1, $3); }
           | infix_expr REAL_DIV      infix_expr
             { $$ = create_expr_infixop($2, e_real_div, $1, $3); }
           | infix_expr STRING_CONCAT infix_expr
             { $$ = create_expr_infixop($2, e_string_concat, $1, $3); }
           | infix_expr PREPEND       infix_expr
             { $$ = create_expr_infixop($2, e_prepend, $1, $3); }
           | infix_expr LIST_CONCAT   infix_expr
             { $$ = create_expr_infixop($2, e_list_concat, $1, $3); }
           ;

applbasicfun: ABS_ BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_abs, $3);
              }
            | MIN_ BRA_L expr COMMA expr BRA_R {
                $$ = create_expr_basicfun2($1, e_min, $3, $5);
              }
            | MAX_ BRA_L expr COMMA expr BRA_R {
                $$ = create_expr_basicfun2($1, e_max, $3, $5);
              }
            | SQRT BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_sqrt, $3);
              }
            | SIN BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_sin, $3);
              }
            | COS BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_cos, $3);
              }
            | EXP BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_exp, $3);
              }
            | LN BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_ln, $3);
              }
            | REAL BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_real, $3);
              }
            | TRUNCATE BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_truncate, $3);
              }
            | PROJECT BRA_L expr COMMA expr BRA_R {
                $$ = create_expr_project($3, $5);
              }
            | SIZE BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_size, $3);
              }
            | ORD BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_ord, $3);
              }
            | CHR BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_chr, $3);
              }
            | SUBSTRING BRA_L expr COMMA expr COMMA expr BRA_R {
                $$ = create_expr_substring($1, $3, $5, $7);
              }
            | NULL_ BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_null, $3);
              }
            | HD BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_hd, $3);
              }
            | TL BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_tl, $3);
              }
            | LENGTH BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_length, $3);
              }
            | REV BRA_L expr BRA_R {
                $$ = create_expr_basicfun($1, e_rev, $3);
              }
            ;

atomic_expr : constant { $$ = $1; }
            | ID { /* may be a variable or a parameterless constructor */
                tmp_pStEntry = seekSymbolId(symbol_table, $1.str); 
                if (tmp_pStEntry != NULL) {
	          //thus id can not be a variable
                  if (tmp_pStEntry->symbol_type == e_st_constructor) {
                    $$ = create_expr_constructor($1, NULL);
		  } else {
                    st_error("id is not a constructor bf", $1);
                    return(1);
		  } //if not a constructor
		} else {
		  //thus id has to be a variable, but is it?
		printf("variable found %s \n",$1.str);
                  tmp_pStEntry = seekSymbolId(binding_table, $1.str);
                  if (tmp_pStEntry != NULL) {
                    $$ = create_expr_variable($1);
		  } else {
printf("add as undef\n");
		    undefined_symbols = addUndefSymbol(undefined_symbols, $1, 0);
		    $$ = create_expr_variable($1);
                    //st_error("id is not a parameter nf", $1);
                    //return(1);
		  }
                  /*# SED-BEGIN delete region #*/
                  /*# SED-END delete region #*/
                  /*# SED-BEGIN insert region #
//                    st_error("id is not a constructor", $1);
  //                  return(1);
                  # SED-END insert region #*/
	        } //if in symbol tabel
	    }
            | BRA_L expr BRA_R            { $$ = $2; }
            | SQRBRA_L SQRBRA_R {
		printf("XXX empty list\n");
	        tmp_mlval.val.mllist = NULL;
		tmp_mlval.lineno = $1.lineno;
		tmp_mlval.colno  = $1.colno;
		//$$ = create_nil(tmp_mlval);
		$$ = create_expr_listkeeper(tmp_mlval, NULL);
	      }
            | SQRBRA_L expr_list SQRBRA_R { 
                //$$ = $2;
		printf("XXX full list\n");
	        tmp_mlval.val.mllist = NULL;
		tmp_mlval.lineno = $1.lineno;
		tmp_mlval.colno  = $1.colno;
                $$ = create_expr_listkeeper(tmp_mlval, $2);
              }
            ;

arg_list : expr {
              $$ = create_expr_arg_list($1, NULL);
            }
         | expr COMMA arg_list {
             $$ = create_expr_arg_list($1, $3);
           }
         ;

expr_list : expr {
              /* tmp_mlval.val.mllist = NULL;
	      tmp_mlval.lineno = -1;
	      tmp_mlval.colno  = -1;
              $$ = create_expr_list($1, create_nil(tmp_mlval));*/
	      $$ = create_expr_list($1, NULL);
            }
          | expr COMMA expr_list {
              $$ = create_expr_list($1, $3);
            }
          ;


%%


void yyyInitialize(void)
{
  DBUG_ENTER("yyyInitialize");

  yyy_InitScanner();
  keep_tree_parts_in_mind = TRUE;
  clear_ErrorDescription();

  /* init garbage collection */
  gcInitialize();

  if (symbol_table == NULL) {
    /* add basic types to symboltable */
    symbol_table=addSymbol(symbol_table,"",      e_st_basictype,0, 0, 0, NULL);    //for dedugging only 
    symbol_table=addSymbol(symbol_table,"unit",  e_st_basictype,0, 1, 0, NULL);
    symbol_table=addSymbol(symbol_table,"bool",  e_st_basictype,0, 2, 0, NULL);
    symbol_table=addSymbol(symbol_table,"int",   e_st_basictype,0, 3, 0, NULL);
    symbol_table=addSymbol(symbol_table,"real",  e_st_basictype,0, 4, 0, NULL);
    symbol_table=addSymbol(symbol_table,"string",e_st_basictype,0, 5, 0, NULL);
    symbol_table=addSymbol(symbol_table,"nil",   e_st_basictype,0, 6, 0, NULL);
    symbol_table=addSymbol(symbol_table,"list",  e_st_basictype,0, 7, 6, NULL);
/*# SED-BEGIN delete region #*/
    next_free_type_index =                                          8;
/*# SED-END delete region #*/
  }

  // clear binding-table for each parse
  /*# SED-BEGIN delete region #*/
  binding_table = NULL;
  /*# SED-END delete region #*/

  //printf_symboltable(symbol_table);

  DBUG_VOID_RETURN;
} //yyyInitialize


void yyyFinalize(void)
{
  DBUG_ENTER("yyyFinalize");

  keep_tree_parts_in_mind = FALSE;
  gcFinalize();

  
  DBUG_VOID_RETURN;
} //yyyFinalize


